# This file was auto-generated by salt's setup on Monday, 09 October 2017 @ 16:10:52 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2016, 11, 8, 0, '', 0, 0, None)
